import os
import pathlib
from typing import Dict

import openapi3.object_base
import requests
import yaml
from dict_tools.data import NamespaceDict


def context(hub, ctx, directory: pathlib.Path):
    if ctx.get("simple_service_name"):
        ctx.service_name = ctx.simple_service_name
    elif not ctx.get("service_name"):
        ctx.service_name = (
            ctx.clean_name.replace("idem", "").replace("cloud", "").strip("_")
        )

    ctx.has_acct_plugin = bool(ctx.acct_plugin)
    if not ctx.has_acct_plugin:
        # Create auth plugins
        ctx.acct_plugin = ctx.service_name

    # Add extra templates provided by openapi3
    if not ctx.get("templates_dir"):
        # Use default templates
        ctx["templates_dir"] = (
            os.path.dirname(os.path.realpath(__file__))
            + "/{{cookiecutter.root_dir}}/{{cookiecutter.clean_name}}/autogen/{{cookiecutter.service_name}}/templates/"
        )

    # all template files are copied without render
    ctx["_copy_without_render"] = ["*.jinja2"]

    merged_plugins = {}
    for specification in ctx.specification:
        # Read the spec from URL or local file (Yaml)
        spec = hub.pop_create.openapi3.init.read(source=specification)
        api = openapi3.OpenAPI(spec, validate=True)
        errors = api.errors()
        if errors:
            for e in errors:
                hub.log.warning(e)

        # list these as defaults in the acct plugin
        if api.servers:
            ctx.servers = [x.url for x in api.servers]
        else:
            ctx.servers = ctx.get("servers", [""])

        hub.log.debug(f"Working with openapi spec version: {api.openapi}")

        ctx.cloud_api_version = api.info.version or "latest"
        ctx.clean_api_version = hub.tool.format.case.snake(ctx.cloud_api_version).strip(
            "_"
        )

        if not ctx.clean_api_version:
            ctx.clean_api_version = ctx.get("clean_api_version")

        if not ctx.cloud_api_version:
            ctx.cloud_api_version = ctx.get("cloud_api_version")

        # If the api version starts with a digit then make sure it can be used for python namespacing
        if ctx.clean_api_version[0].isdigit():
            ctx.clean_api_version = "v" + ctx.clean_api_version

        # Get function plugins
        plugins = hub.pop_create.openapi3.parse.plugins(ctx, api)

        # Add any missing function which is required for idem resource modules
        plugins = hub.pop_create.openapi3.init.add_missing_known_functions(ctx, plugins)

        # Add tests function
        plugins = hub.pop_create.openapi3.init.add_tests_functions(ctx, plugins)
        merged_plugins.update(plugins)

    ctx.cloud_spec = NamespaceDict(
        api_version=ctx.cloud_api_version,
        project_name=ctx.project_name,
        service_name=ctx.service_name,
        request_format={},
        plugins=merged_plugins,
    )

    hub.pop_create.init.run(directory=directory, subparsers=["idem_cloud"], **ctx)

    return ctx


def read(hub, source: str or Dict):
    """
    If the path is a file, then parse the json contents of the file,
    If the path is a url, then return a json response from the url.
    """
    if isinstance(source, Dict):
        return source

    path = pathlib.Path(source)

    if path.exists():
        with path.open("r") as fh:
            ret = yaml.safe_load(fh)
    else:
        request = requests.get(source, headers={"Content-Type": "application/json"})
        ret = request.json()

    return ret


def add_missing_known_functions(hub, ctx, plugins):
    # This is to make sure we still create standard skeletons for CRUD operations in exec
    # This way the only missing part would be API call in those modules
    for ref in list(plugins):
        for idem_func_name in ["get", "list", "create", "update", "delete"]:
            if not plugins[ref].get("functions", {}).get(idem_func_name):
                plugins[ref]["functions"][idem_func_name] = {
                    "doc": "",
                    "params": {},
                    "hardcoded": {
                        "method": "TODO",
                        "path": "TODO",
                        "service_name": ctx.service_name,
                        "resource_name": ref,
                    },
                }

    if ctx.create_plugin == "state_modules":
        for ref in list(plugins):
            for func_name in list(plugins.get(ref).get("functions", {})):
                if func_name == "create":
                    plugins[ref]["functions"]["present"] = (
                        plugins.get(ref).get("functions", {}).get(func_name)
                    )
                elif func_name == "delete":
                    plugins[ref]["functions"]["absent"] = (
                        plugins.get(ref).get("functions", {}).get(func_name)
                    )
                elif func_name == "list":
                    plugins[ref]["functions"]["describe"] = (
                        plugins.get(ref).get("functions", {}).get(func_name)
                    )

    return plugins


def add_tests_functions(hub, ctx, plugins):
    for ref in list(plugins):
        functions = plugins[ref].get("functions", {}).copy()
        for func_name, func_data in functions.items():
            plugins[ref]["functions"][
                f"test_{func_name}"
            ] = hub.pop_create.openapi3.function.create_test_for_function(
                ctx=ctx, func_name=func_name, func_data=func_data
            )

    return plugins



import os
import pathlib

from cloudspec import CloudSpec


def run(hub, ctx, root_directory: pathlib.Path or str):
    if isinstance(root_directory, str):
        root_directory = pathlib.Path(root_directory)
    cloud_spec = CloudSpec(**ctx.cloud_spec)
    exec_dir = root_directory / ctx.clean_name / "exec" / ctx.service_name
    template_dir = (
        root_directory
        / ctx.clean_name
        / "autogen"
        / ctx.service_name
        / "templates"
        / "exec"
    )

    for ref, plugin in cloud_spec.plugins.items():
        mod_file = hub.cloudspec.parse.plugin.touch(exec_dir, ref)
        ref = hub.cloudspec.parse.plugin.ref(ctx, ref)
        cli_ref = hub.cloudspec.parse.plugin.mod_ref(ctx, ref, plugin)

        plugin["func_alias"] = {"list_": "list"}

        to_write = hub.cloudspec.parse.plugin.header(plugin)

        for function_name, function_data in plugin.functions.items():
            if function_name in ["present", "absent", "describe"]:
                # If plugin comes in with all methods, let's not create for state_modules functions
                continue

            request_format_template_file = f"{template_dir}/{function_name}.jinja2"
            if os.path.isfile(request_format_template_file):
                with open(f"{template_dir}/{function_name}.jinja2", "rb+") as f:
                    request_format = f.read().decode()
            else:
                request_format = cloud_spec.request_format.get(function_name)

            template = hub.tool.jinja.template(
                f"{hub.cloudspec.template.exec.FUNCTION}\n    {request_format}\n\n\n"
            )

            function_alias = plugin.func_alias.get(function_name, function_name)

            try:
                to_write += template.render(
                    function=dict(
                        name=function_name,
                        hardcoded=function_data.hardcoded,
                        doc=hub.cloudspec.parse.function.doc(function_data)
                        + hub.cloudspec.parse.param.sphinx_docs(function_data.params)
                        + hub.cloudspec.parse.function.return_type(function_data),
                        ref=f"{ref}.{function_alias}",
                        cli_ref=f"{cli_ref}.{function_alias}",
                        header_params=hub.cloudspec.parse.param.headers(
                            function_data.params
                        ),
                        required_call_params=hub.cloudspec.parse.param.required_call_params(
                            function_data.params
                        ),
                        return_type=function_data.return_type,
                    ),
                    parameter=dict(
                        mapping=hub.cloudspec.parse.param.mappings(function_data.params)
                    ),
                )
            except Exception as err:
                hub.log.warning(
                    f"Warning when generating function's action definitions for {function_name}: {err.__class__.__name__}: {err}"
                )

        mod_file.write_text(to_write)



import jinja2


def template(hub, data: str) -> jinja2.Template:
    """
    return a rendered template from a string the same way every time
    """
    # noinspection JinjaAutoinspect
    return jinja2.Environment(  # nosec
        autoescape=False, lstrip_blocks=True, enable_async=True
    ).from_string(data)



import pathlib
import shutil

import pop.hub
from dict_tools.data import NamespaceDict

if __name__ == "__main__":
    root_directory = pathlib.Path.cwd()

    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="pop_create")
    hub.pop.sub.add(dyne_name="config")
    hub.config.integrate.load(["pop_create"], "pop_create", parse_cli=False, logs=False)
    ctx = NamespaceDict({{cookiecutter}})

    idem_dynes = ["exec", "states", "tool"]
    if not ctx.has_acct_plugin:
        idem_dynes.append("acct")

    dyne_list = set(idem_dynes + list(ctx.short_dyne_list))
    # FIXME generating tests is problematic for some reason
    # pop_create_core = ["seed", "tests", "docs"]
    pop_create_core = ["seed", "docs"]
    seed_ctx = NamespaceDict(vertical=True, dyne_list=dyne_list)

    if "cloud_spec" not in ctx:
        # The openapi3/swagger plugin might call this with their own spec
        ctx.cloud_spec = {}

    if ctx.get("templates_dir", None):
        target = (
            root_directory / ctx.clean_name / "autogen" / ctx.service_name / "templates"
        )
        shutil.copytree(ctx.get("templates_dir"), target, dirs_exist_ok=True)

    # Bring in the full context, but don't override any of our overrides
    for k in ctx:
        if k not in seed_ctx:
            seed_ctx[k] = ctx[k]

    # Run a traditional pop seed on the directory with common idem states
    hub.pop_create.init.run(
        directory=root_directory, subparsers=pop_create_core, **seed_ctx
    )



import re


def camel(hub, string: str, dromedary: bool = False) -> str:
    """
    Change a snake-cased string to a camel-cased string
    """
    if "_" not in string and (dromedary or string[0].isupper()):
        # Already cameled
        return string
    else:
        # Replace underscores with spaces then call str()'s title() method, then get rid of spaces
        result = string.replace("_", " ").title().replace(" ", "")
        if dromedary:
            return result[0].swapcase() + result[1:]
        return result


def snake(hub, string: str) -> str:
    """
    Change a camel-cased string to a snake-cased string
    """
    # Separate each camel-cased word into underscore delimited words
    string = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", string)
    string = re.sub("([a-z0-9])([A-Z])", r"\1_\2", string)
    # Replace special characters with underscores
    string = re.sub(r"[^\w]", "_", string)
    # make sure everything is lower-cased
    string = string.lower()
    string = string.replace("__", "_")

    return string



"""
Templates for state modules
"""

PRESENT_FUNCTION = """
async def {{function.name}}(
    hub,
    ctx,
    name: str
    {% if function.header_params %}
        {{function.header_params}}
    {% endif -%},
    resource_id: str = None,
    **kwargs
) -> Dict[str, Any]:
    r'''
    **Autogenerated function**
    {{function.doc|replace("'" * 3, '"' * 3)}}
    Examples:

        .. code-block:: sls

            resource_is_{{function.name}}:
              {{ function.ref }}.{{ function.name }}:
                - {{ function.required_call_params.split(", ")|join("\n                - ")|replace("=", ": ") }}
    '''
"""

ABSENT_FUNCTION = """
async def {{function.name}}(
    hub,
    ctx,
    name: str
    {% if function.header_params %}
        {{function.header_params}}
    {% endif -%},
    resource_id: str = None,
    **kwargs
)  -> Dict[str, Any]:
    r'''
    **Autogenerated function**
    {{function.doc|replace("'" * 3, '"' * 3)}}
    Examples:

        .. code-block:: sls

            resource_is_{{function.name}}:
              {{ function.ref }}.{{ function.name }}:
                - {{ function.required_call_params.split(", ")|join("\n                - ")|replace("=", ": ") }}
    '''
"""

DESCRIBE_FUNCTION = """
async def {{function.name}}(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r'''
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    {{function.doc|replace("'" * 3, '"' * 3)}}
    Examples:

        .. code-block:: bash

            $ idem describe {{ function.ref }}
    '''
"""



import os
import pathlib

from cloudspec import CloudSpec
from cloudspec import CloudSpecParam


def run(hub, ctx, root_directory: pathlib.Path or str):
    if isinstance(root_directory, str):
        root_directory = pathlib.Path(root_directory)
    cloud_spec = CloudSpec(**ctx.cloud_spec)

    for ref, plugin in cloud_spec.plugins.items():
        plugin["imports"] = ["import pytest"]

        for test_type in ["exec", "states", "tool"]:
            test_dir = (
                root_directory
                / "tests"
                / "integration"
                # test_module_type = exec, states, tool
                / test_type
            )

            template_dir = (
                root_directory
                / ctx.clean_name
                / "autogen"
                / ctx.service_name
                / "templates"
                / "tests"
                # test_module_type = exec, states, tool
                / test_type
            )

            mod_file = hub.cloudspec.parse.plugin.touch(test_dir, f"test_{ref}")

            plugin["file_vars"] = {}
            if test_type == "states":
                # The states test should be run with --test and actual run mode
                plugin["file_vars"]["PARAMETRIZE"] = {
                    "argnames": "__test",
                    "argvalues": [True, False],
                    "ids": ["--test", "run"],
                }

            if test_type in ["states", "exec"]:
                # Add default global parameter for tracking a resource being created inside test
                # This helps in clean up of a created resource in test
                plugin["file_vars"]["PARAMETER"] = {
                    "name": "idem-test-resource- + TODO: Add unique identifier generator",
                }

            to_write = hub.cloudspec.parse.plugin.header(plugin)

            # Collect get_call_params beforehand from the list
            # This is helpful when a test method wants to call into GET after any operation in a test
            get_call_params = None
            test_get_func = [
                value for key, value in plugin.functions.items() if key == "test_get"
            ]
            if test_get_func:
                get_call_params = {
                    name: CloudSpecParam(name=name, **param_spec)
                    for name, param_spec in test_get_func[0]
                    .get("hardcoded", {})
                    .get("method_call_params", {})
                    .items()
                }

            for function_name, function_data in plugin.functions.items():
                if not function_name.startswith("test_"):
                    # Only process test functions
                    continue

                if test_type != function_data.hardcoded.test_module_type:
                    # Let's not add to the file if it doesn't belong to the module
                    continue

                request_format_template_file = f"{template_dir}/{function_name}.jinja2"
                default_request_format_template_file = f"{template_dir}/default.jinja2"
                if os.path.isfile(request_format_template_file):
                    with open(f"{template_dir}/{function_name}.jinja2", "rb+") as f:
                        request_format = f.read().decode()
                elif os.path.isfile(default_request_format_template_file):
                    with open(default_request_format_template_file, "rb+") as f:
                        request_format = f.read().decode()
                else:
                    request_format = cloud_spec.request_format.get(function_name)

                template = hub.tool.jinja.template(
                    f"{hub.cloudspec.template.test.TEST_FUNCTION}\n    {request_format}\n\n\n"
                )

                # Get call params for method's (being called in the test)
                method_call_params = {
                    name: CloudSpecParam(name=name, **param_spec)
                    for name, param_spec in function_data.hardcoded.method_call_params.items()
                }

                try:
                    method_call_name = function_data.hardcoded.method_call_name
                    to_write += template.render(
                        function={
                            "name": function_name,
                            "service_name": cloud_spec.service_name,
                            "hardcoded": {
                                "parametrize": True
                                if method_call_name in ["present", "absent"]
                                else False,
                                **function_data.hardcoded,
                            },
                            "method_call_name": method_call_name,
                            "method_call_params": hub.cloudspec.parse.param.all_call_params(
                                method_call_params
                            ),
                            "get_call_params": hub.cloudspec.parse.param.all_call_params(
                                get_call_params
                            )
                            if get_call_params is not None
                            else "",
                        },
                    )
                except Exception as err:
                    hub.log.warning(
                        f"Warning when generating function's action definitions for {function_name}: {err.__class__.__name__}: {err}"
                    )
            mod_file.write_text(to_write)



from typing import Dict
from typing import List

from cloudspec import CloudSpecFunction


def return_type(hub, func_data: CloudSpecFunction):
    ret = ""
    if func_data.return_type:
        ret += "\n\n    Returns:\n"
        ret += f"        {func_data.return_type}"
        ret += "\n"
    return ret


def doc(hub, func_data: CloudSpecFunction):
    if func_data.doc:
        doc = func_data.doc.replace('"""', "'''")
        doc = "\n" + hub.tool.format.wrap.indent(doc, 1) + "\n"
    else:
        doc = ""
    return doc


def parse(hub, functions: Dict[str, CloudSpecFunction], targets: List[str]):
    funcs = {}
    if not functions:
        # It might be an init file
        return funcs
    for function_name in targets:
        function_data = functions.get(function_name)
        # This lets us create partial plugin
        if function_data:
            doc = hub.cloudspec.parse.function.doc(function_data)
            doc += hub.cloudspec.parse.param.sphinx_docs(function_data.params)
            doc += "\n\n    Returns:\n        Dict[str, Any]\n"

            if "absent" == function_name and function_data.params:
                # By idem's contract,
                #   - all absent function parameters except 'name' should not be required
                for param_name, metadata in function_data.params.items():
                    if param_name.lower() != "name":
                        metadata["required"] = False

            param_mapping = hub.cloudspec.parse.param.mappings(function_data.params)

            funcs[function_name] = dict(
                function=dict(
                    name=function_name,
                    hardcoded=function_data.hardcoded,
                    doc=doc,
                    header_params=hub.cloudspec.parse.param.headers(
                        function_data.params
                    ),
                    required_call_params=hub.cloudspec.parse.param.required_call_params(
                        function_data.params
                    ),
                ),
                parameter=dict(
                    mapping=param_mapping,
                ),
            )
    return funcs



CLI_CONFIG = {
    "input_file": {"positional": True, "subcommands": ["validate", "create"]},
    "create_plugins": {"options": ["--plugin", "-P"], "subcommands": ["create"]},
    "output": {"source": "rend", "subcommands": ["validate"]},
    "directory": {
        "options": ["-D"],
        "subcommands": ["create_plugins"],
        "source": "pop_create",
    },
}
CONFIG = {
    "input_file": {
        "default": None,
        "help": "A json file to validate for idem-cloud cloudspec format",
    },
    "create_plugins": {
        "default": [],
        "help": "cloudspec create plugins to run against the spec",
    },
}
SUBCOMMANDS = {
    "validate": {},
    "create": {},
}
DYNE = {"pop_create_idem": ["cloudspec"]}



import os
import pathlib

from cloudspec import CloudSpec


def run(hub, ctx, root_directory: pathlib.Path or str):
    if isinstance(root_directory, str):
        root_directory = pathlib.Path(root_directory)
    cloud_spec = CloudSpec(**ctx.cloud_spec)
    states_dir = root_directory / ctx.clean_name / "states" / ctx.service_name
    template_dir = (
        root_directory
        / ctx.clean_name
        / "autogen"
        / ctx.service_name
        / "templates"
        / "state"
    )

    for ref, plugin in cloud_spec.plugins.items():
        mod_file = hub.cloudspec.parse.plugin.touch(states_dir, ref)
        ref = hub.cloudspec.parse.plugin.ref(ctx, ref)
        state_ref = hub.cloudspec.parse.plugin.mod_ref(ctx, ref, plugin)

        plugin["contracts"] = ["resource"]

        to_write = hub.cloudspec.parse.plugin.header(plugin)
        if not plugin.functions:
            mod_file.write_text(to_write)
            continue
        func_data = hub.cloudspec.parse.function.parse(
            plugin.functions, targets=("present", "absent", "describe")
        )

        present_parameter = {}
        if func_data.get("present"):
            present_parameter = hub.cloudspec.parse.param.simple_map(
                plugin.functions["present"].params
            )

        common_params = {}
        if func_data.get("absent") and func_data.get("describe"):
            for name, param in plugin.functions["absent"].params.items():
                if name in plugin.functions["describe"].params:
                    common_params[name] = plugin.functions["describe"].params[name]

            common_params = hub.cloudspec.parse.param.mappings(common_params)

        # Create the present, absent, and describe functions; these are required for every state module
        for function_name, TEMPLATE in zip(
            ("present", "absent", "describe"),
            (
                hub.cloudspec.template.state.PRESENT_FUNCTION,
                hub.cloudspec.template.state.ABSENT_FUNCTION,
                hub.cloudspec.template.state.DESCRIBE_FUNCTION,
            ),
        ):
            if func_data.get(function_name):
                request_format_template_file = f"{template_dir}/{function_name}.jinja2"
                if os.path.isfile(request_format_template_file):
                    with open(f"{template_dir}/{function_name}.jinja2", "rb+") as f:
                        request_format = f.read().decode()
                else:
                    request_format = cloud_spec.request_format.get(function_name)

                template = hub.tool.jinja.template(
                    f"{TEMPLATE}\n    {request_format}\n\n\n"
                )

                try:
                    to_write += template.render(
                        service_name=cloud_spec.service_name,
                        function=dict(
                            ref=ref,
                            state_ref=f"states.{state_ref}",
                            **func_data[function_name]["function"],
                        ),
                        parameter=func_data[function_name]["parameter"],
                        present_parameter=present_parameter,
                        get_params=common_params,
                    )
                except Exception as err:
                    hub.log.warning(
                        f"Warning when generating function's action definitions for {function_name}: {err.__class__.__name__}: {err}"
                    )

                mod_file.write_text(to_write)



from typing import Dict

from cloudspec import CloudSpecParam


def sphinx_docs(hub, parameters: CloudSpecParam) -> str:
    """
    Get the sphinx docs for the parameters
    """
    if not parameters:
        return ""

    ret = "\n    Args:\n"
    for param in parameters.values():
        ret += f"        {param.snaked}"
        # Doc strings standard is lower case list[]
        if param.param_type is None:
            param_type = None
        else:
            param_type = param.param_type.replace("List[", "list[")
        if param.member:
            # Complex params are represented as Dict[str, Any] in the sphinx docs
            param_type = param_type.format("Dict[str, Any]")

        if param_type:
            ret += f"({param_type}{', optional' if not param.required else ''})"

        ret += f": {param.doc}"
        if not ret.endswith("."):
            ret += "."
        if not param.required:
            ret += f" Defaults to {param.default}."

        ret += nested_params_docs(param, "            ")
        ret += "\n"

    return ret.rstrip()


def nested_params_docs(param: CloudSpecParam, offset: str):
    """
    Get the sphinx docs for the nested parameters
    """
    ret = ""
    if param.member:
        for nested_param_name, nested_param_data in param.member.params.items():
            # Doc strings standard is lower case list[]
            param_type = nested_param_data.param_type.replace("List[", "list[")
            if nested_param_data.member:
                # This is the next level of nested complex argument.
                param_type = param_type.format("Dict[str, Any]")
            ret += f"\n{offset}* "
            ret += f"{nested_param_name} ({param_type}{', optional' if not nested_param_data.required else ''}): {nested_param_data.doc}"

            # Recursively add doc for all nested complex arguments
            ret += nested_params_docs(nested_param_data, offset + "    ")
    return ret


def headers(hub, parameters: CloudSpecParam) -> str:
    """
    The arguments that will be put in the function definition
    """
    ret = ""

    required_params = {
        name: data for name, data in parameters.items() if data["required"]
    }
    for param_name, param_data in required_params.items():
        ret += add_param_to_header(param_data)

    unrequired_params = {
        name: data for name, data in parameters.items() if not data["required"]
    }

    for param_name, param_data in unrequired_params.items():
        ret += add_param_to_header(param_data)

        # TODO handle this case properly
        if param_data.default == "{}" or param_data.default == "[]":
            ret += f" = None"
        else:
            ret += f" = {param_data.default}"

    return ret


def add_param_to_header(param_data: Dict) -> str:
    ret = f",\n    {param_data.snaked}: "
    if param_data.member:
        # Insert nested complex format for as the argument type.
        # param_type should be "{}|List[{}]|Dict[str, {}]"
        ret += f"{param_data.param_type}".format(
            add_nested_param_to_header(param_data.member, "        ")
        )
    else:
        ret += f"{param_data.param_type}"

    return ret


def add_nested_param_to_header(member: Dict, offset: str) -> str:
    """
    Create a dataclass for a complex nested argument
    """
    ret = f'make_dataclass(\n{offset}"{member.name}",\n{offset}['
    required_params = {
        name: data for name, data in member.get("params").items() if data["required"]
    }

    for param_name, param_data in required_params.items():
        ret += add_nested_param(param_name, param_data, offset)

    unrequired_params = {
        name: data
        for name, data in member.get("params").items()
        if not data["required"]
    }

    for param_name, param_data in unrequired_params.items():
        ret += add_nested_param(param_name, param_data, offset)

    # Remove last ', '
    ret = ret[0:-2]
    ret += f"\n{offset}]\n{offset})"

    return ret


def add_nested_param(param_name: str, param_data, offset: str):
    ret = ""
    ret += f"\n{offset}    "
    # TODO: Nested param name should be snaked
    ret += f'("{param_name}", '
    if param_data.member:
        # Recursively add all nested complex arguments
        ret += param_data.param_type.format(
            add_nested_param_to_header(param_data.member, offset + "    ")
        )
    else:
        ret += f"{param_data.param_type}"
    if param_data["required"]:
        ret += f"), "
    else:
        ret += f", field(default={param_data.default})), "

    return ret


def required_call_params(hub, parameters: CloudSpecParam) -> str:
    """
    Get a mapping of required function args to the values that will be used in the final caller
    """
    ret = []

    required_params = {
        name: data for name, data in parameters.items() if data["required"]
    }
    for param_data in required_params.values():
        ret.append(f"{param_data.snaked}={param_data.default or 'value'}")

    return ", ".join(ret)


def all_call_params(hub, parameters: CloudSpecParam) -> str:
    """
    Get a mapping of all function args to the values that will be used in the final caller
    """
    ret = []

    all_params = {name: data for name, data in parameters.items()}
    for param_data in all_params.values():
        ret.append(f"{param_data.snaked}={param_data.default or 'value'}")

    return ", ".join(ret)


def mappings(hub, parameters: CloudSpecParam) -> Dict[str, str]:
    ret = {}
    map_params = {
        name: data
        for name, data in parameters.items()
        if data["target_type"] == "mapping"
    }
    for name, data in map_params.items():
        target = data["target"]
        if target not in ret:
            ret[target] = {}
        ret[target][name] = data.snaked

    fmt = lambda item: ", ".join(f'"{k}": {v}' for k, v in item.items())
    return {k: f"{{{fmt(v)}}}" for k, v in ret.items()}


def simple_map(hub, parameters: Dict[str, CloudSpecParam]):
    result = {}
    for k, param_data in parameters.items():
        if k == "name" and param_data.target_type == "arg":
            continue
        result[param_data.snaked] = k
    return result



import shutil
import subprocess


COMMAND = shutil.which("gradle")


def __virtual__(hub):
    if not COMMAND:
        return False, "gradle command not found"
    return (True,)


def run(hub, subcommand: str, project_dir: str, *args):
    proc = subprocess.Popen(
        [COMMAND, subcommand, f"--project-dir={project_dir}", *args],
        stderr=subprocess.PIPE,
        cwd=project_dir,
        encoding="ascii",
    )
    _, stderr = proc.communicate()
    for line in stderr.splitlines():
        hub.log.error(line)
    code = proc.wait()
    if code:
        raise ChildProcessError(f"Gradle command failed with status {code}")


def clean(hub, build_dir: str):
    hub.tool.gradle.run("clean", project_dir=build_dir)


def build(hub, build_dir: str):
    hub.tool.gradle.run("build", project_dir=build_dir)

    # Return the resulting openapi spec path
    return {}



TEST_FUNCTION = """
@pytest.mark.asyncio
{% if function.hardcoded.parametrize %}@pytest.mark.parametrize(**PARAMETRIZE){% endif %}
async def {{function.name}}(hub, ctx{% if function.hardcoded.parametrize %}, __test{% endif %}):
    r'''
    **Autogenerated test function**
    '''
"""



def __init__(hub):
    hub.exec.{{cookiecutter.service_name}}.ENDPOINT_URLS = {{cookiecutter.servers}}
    # The default is the first in the list
    hub.exec.{{cookiecutter.service_name}}.DEFAULT_ENDPOINT_URL = "{{cookiecutter.servers|first}}"

    # This enables acct profiles that begin with "{{cookiecutter.acct_plugin}}" for {{cookiecutter.service_name}} modules
    hub.exec.{{cookiecutter.service_name}}.ACCT = ["{{cookiecutter.acct_plugin}}"]



from typing import Any
from typing import Dict

import openapi3.object_base
import tqdm


def plugins(hub, ctx, api: openapi3.OpenAPI) -> Dict[str, Any]:
    ret = {}
    paths: openapi3.object_base.Map = api.paths

    all_schemas = hub.pop_create.openapi3.schemas.parse(api)

    for name, path in tqdm.tqdm(paths.items(), desc="Parsing paths"):
        if not isinstance(path, openapi3.paths.Path):
            # Let's not fail but continue to other paths instead
            hub.log.warning(
                f"The {name} is not an instance of Path. It will not be parsed."
            )
            continue

        # Get the request type that works for this request
        for request_type in path.raw_element.keys():
            func: openapi3.paths.Operation = getattr(path, request_type)
            if not func:
                continue

            plugin = None
            refs = None
            if func.tags:
                subs = [hub.tool.format.case.snake(sub) for sub in func.tags]
                if not subs:
                    plugin = "init"
                else:
                    plugin = subs.pop()

                refs = subs + [plugin]
            elif func.get_path():
                # e.g. "/pets", /pets/{id} -> "pets" become plugin
                # func.get_path() could look like this "paths./pets/{id}.get
                # second element is always the actual path itself
                func_path = func.get_path().split(".")[1]
                plugin = func_path.lstrip("/").split("/")[0]
                refs = [plugin]

            if not plugin:
                raise AttributeError(
                    f"Not sure how to find plugin for {name}. Try using 'tags'"
                )

            # service name is already added in pop-create
            ref = ".".join(refs)
            if ref not in ret:
                # This is the first time we have looked at this plugin
                ret[ref] = {
                    "functions": {},
                    "doc": "",
                    "imports": [
                        "from typing import Any",
                        "from typing import Dict",
                        "from typing import List",
                        "from collections import OrderedDict",
                        "from dataclasses import field",
                        "from dataclasses import make_dataclass",
                        "import dict_tools.differ as differ",
                    ],
                }

            # See if this function will be reserved CRUD operations, if so change the name
            reserved_func_name = (
                hub.pop_create.openapi3.parse.resolve_reserved_function_name(
                    name, plugin, request_type, func
                )
            )

            # Anything that doesn't resolved in reserved function name, would go into tools/{resource_name}/*
            func_name = (
                hub.pop_create.openapi3.parse.resolve_function_name(
                    name, plugin, request_type, func
                )
                if not reserved_func_name
                else reserved_func_name
            )

            try:
                ret[ref]["functions"][
                    func_name
                ] = hub.pop_create.openapi3.function.parse(
                    ctx,
                    request_type,
                    name,
                    plugin,
                    func,
                    all_schemas,
                )
            except Exception as e:
                hub.log.debug(
                    f"Failed to parse function data for {plugin}: {request_type} '{name}': {e.__class__.__name__}: {e}"
                )

    return ret


def resolve_reserved_function_name(
    hub, name: str, plugin: str, request_type: str, func: openapi3.paths.Operation
):
    # Do not make deprecated API paths into CRUDs and throw them into tools/*
    possible_resource_names = tuple(
        [
            # e.g. /store_order, /store-order, /store_orders, /store-orders
            plugin,
            plugin.replace("_", "-"),
            f"{plugin}s",
            f"{plugin.replace('_', '-')}s",
        ]
    )
    if not func.deprecated:
        # e.g. plugin: pets, path: /pets
        if name.endswith(possible_resource_names):
            # list/post
            if request_type == "get":
                return "list"
            elif request_type == "post":
                return "create"
            elif request_type == "put":
                # PUT /pets
                return "update"
        # e.g.: plugin: pets, path: /pets/{id}
        elif name.rsplit("/", 1)[0].endswith(possible_resource_names):
            # get/list/put
            if request_type == "get":
                return "get"
            elif request_type == "put" or request_type == "patch":
                return "update"
            elif request_type == "delete":
                return "delete"

    return None


def resolve_function_name(
    hub, name: str, plugin: str, request_type: str, func: openapi3.paths.Operation
):
    # This is the preferred way to get a function name
    # However, some APIs can just put reserved/known operationId which is not helpful here in parsing and codegen
    # Also, we always handle reserved/known operation names in different way
    func_name = (
        func.operationId
        if func.operationId
        not in ["get", "list", "create", "patch", "update", "put", "delete"]
        else None
    )

    # Fallback function name based on the pets example
    if not func_name and " " in name:
        func_name = "_".join(name.split(" ")[1:]).lower()

    if not func_name and func.summary:
        # Truncate it to 10 characters and add plugin name in the end
        func_name = f"{func.summary[:10]}_{plugin}"

    if not func_name and request_type:
        # No operationId, no summary, let's create it with path
        # GET /pets/{id}/types -> get_pets_id_types
        func_name = (
            f"{request_type}_{name.replace('{', '').replace('}', '').replace('/', '_')}"
        )

    if not func_name and func.extensions:
        func_name = func.extensions[sorted(func.extensions.keys())[0]]

    # Maybe we need more fallbacks, you tell me
    if not func_name:
        # Maybe a fallback based on the path and method?
        raise AttributeError(f"Not sure how to find func name for {name}, help me out")

    return hub.tool.format.case.snake(func_name)



"""Add test related fixtures here e.g. a fixture that will pre-create a resource needed for the tests."""


import asyncio
import pathlib
from typing import Dict, Any, List

import dict_tools
import pytest_asyncio

import pop

import pytest


@pytest.fixture(scope="session")
def code_dir() -> pathlib.Path:
    print(f"code_dir:{pathlib.Path(__file__).parent.parent.absolute()}")
    return pathlib.Path(__file__).parent.parent.absolute()


@pytest.fixture(scope="module")
def acct_data(ctx):
    """
    acct_data that can be used in running simple yaml blocks
    """
    yield {"profiles": {"{{cookiecutter.service_name}}": {"default": ctx.acct}}}


@pytest.fixture(scope="module", autouse=True)
def acct_profile() -> str:
    return "TODO: Add profile name to use for test"


@pytest.fixture(scope="module")
def event_loop():
    loop = asyncio.get_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="session")
def acct_subs():
    return ["{{cookiecutter.service_name}}"]


@pytest.fixture(scope="module", name="hub")
def integration_hub(code_dir, event_loop):
    hub = pop.hub.Hub()
    hub.pop.loop.CURRENT_LOOP = event_loop
    hub.pop.sub.add(dyne_name="idem")
    hub.pop.config.load(hub.idem.CONFIG_LOAD, "idem", parse_cli=False)
    hub.idem.RUNS = {"test": {}}

    yield hub


@pytest_asyncio.fixture(scope="module", name="ctx")
async def integration_ctx(
    hub, acct_subs: List[str], acct_profile: str
) -> Dict[str, Any]:
    ctx = dict_tools.data.NamespaceDict(
        run_name="test",
        test=False,
        tag="fake_|-test_|-tag",
        old_state={},
    )

    if not hub.OPT.acct.acct_file:
        raise ConnectionError("No ACCT_FILE in the environment")
    if not hub.OPT.acct.acct_key:
        raise ConnectionError("No ACCT_KEY in the environment")

    await hub.acct.init.unlock(hub.OPT.acct.acct_file, hub.OPT.acct.acct_key)
    ctx.acct = await hub.acct.init.gather(acct_subs, acct_profile)
    if not ctx.acct:
        raise Exception(
            f'Unable to load acct "{acct_profile}" from "{hub.OPT.acct.acct_file}"'
        )

    yield ctx


@pytest.fixture(
    scope="function",
    name="__test",
    params=[0, 1, 2, 3],
    ids=["--test", "run", "no change --test", "no change"],
)
def test_flag(ctx, request):
    """
    Functions that use the __test fixture will be run four times
    0. With the --test flag set, to test what happens before creating a resource
    1. Without the --test flag set, to create a resource
    2. With the --test flag set, to verify that no changes would be made
    3. Without the --test flag set, to verify that no changes are made
    """
    ctx.test = bool((request.param + 1) % 2)
    yield request.param



"""Parse function related data from swagger"""
from typing import Any
from typing import Dict

import openapi3


def parse(
    hub,
    ctx,
    request_type: str,
    path: str,
    resource_name: str,
    func: openapi3.paths.Operation,
    all_schemas: dict,
) -> Dict[str, Any]:
    params = {}

    # Parse all query, path, header parameters
    for p in func.parameters:
        params[p.name] = hub.pop_create.openapi3.params.parse_non_schema_members(p)

    # Parse request body if available
    (
        request_body_params,
        request_mappings,
    ) = hub.pop_create.openapi3.function.parse_request_body(func, all_schemas)
    params.update(request_body_params)

    if "name" in params:
        # The template already adds name by default
        params.pop("name")

    deprecated_text = "\nDEPRECATED" if func.deprecated else ""

    return {
        "doc": f"{func.summary}\n    {func.description}    {deprecated_text}".strip(),
        "params": params,
        "hardcoded": {
            "method": request_type,
            "path": path.split(" ")[0],
            "service_name": ctx.service_name,
            "resource_name": resource_name,
            "request_mappings": request_mappings,
            "response_mappings": hub.pop_create.openapi3.function.parse_response_mappings(
                func, all_schemas
            ),
        },
    }


def parse_request_body(
    hub,
    func: openapi3.paths.Operation,
    all_schemas: dict,
):
    # Usually put, post, patch should have request_body
    request_body_params = {}
    request_mappings = ""
    request_body: openapi3.paths.RequestBody = func.requestBody
    try:
        if request_body:
            # application/json is preferable always
            # An example would be ->
            #   "application/json": {
            #       "schema": {
            #         "$ref": "#/components/schemas/User"
            #       }
            #   }
            schema = request_body.content.raw_element.get("application/json")
            if not schema:
                # Take the first one for now
                # e.g. #/components/schemas/StoreOrder
                schema = list(request_body.content.raw_element.values()).pop()

            schema_ref = schema.get("schema", {}).get("$ref")
            hub.pop_create.openapi3.params.parse_schema_members(
                request_body_params, schema_ref, all_schemas
            )

            request_mappings = (
                hub.pop_create.openapi3.params.parse_resource_to_request_input_mappings(
                    schema_ref, all_schemas
                )
            )
    except Exception as e:
        hub.log.debug(f"Failed to parse request body: {e.__class__.__name__}: {e}")

    return request_body_params, request_mappings


def parse_response_mappings(
    hub,
    func: openapi3.paths.Operation,
    all_schemas: dict,
):
    responses: openapi3.paths.Response = func.responses
    response_mappings = {}
    try:
        if responses:
            for status, response in responses.raw_element.items():
                # For now, we support 200 only for mapping into "present" format
                if status == "200" and "content" in response:
                    schema = response.get("content", {}).get("application/json")
                    if not schema:
                        # Take the first one for now
                        schema = list(response.get("content", {}).values()).pop()

                    schema_ref = schema.get("schema", {}).get("$ref")

                    response_mappings = (
                        hub.pop_create.openapi3.params.parse_response_mappings(
                            schema_ref, all_schemas
                        )
                    )
    except Exception as e:
        hub.log.debug(f"Failed to parse response mapping: {e.__class__.__name__}: {e}")

    return response_mappings


def create_test_for_function(
    hub, ctx, func_name: str, func_data: dict
) -> Dict[str, Any]:
    if func_name in ["get", "list", "create", "update", "delete"]:
        test_module_type = "exec"
    elif func_name in ["present", "absent", "describe"]:
        test_module_type = "states"
    else:
        test_module_type = "tool"

    return {
        "doc": "",
        "hardcoded": {
            "resource_name": func_data.get("hardcoded", {}).get("resource_name"),
            "service_name": ctx.service_name,
            # This is critical to derive rendering location
            "test_module_type": test_module_type,
            # Add calling method's metadata such as name, call params etc.
            "method_call_name": func_name,
            "method_call_params": func_data.get("params"),
        },
    }



from typing import Any
from typing import Dict
from typing import List
from unittest import mock

import pytest
from dict_tools.data import NamespaceDict


# ================================================================================
# pop fixtures
# ================================================================================
@pytest.fixture(scope="session", autouse=True)
def acct_subs() -> List[str]:
    return ["{{cookiecutter.service_name}}"]


@pytest.fixture(scope="session", autouse=True)
def acct_profile() -> str:
    # The name of the profile that will be used for tests
    # Do NOT use production credentials for tests!
    # A profile with this name needs to be defined in the $ACCT_FILE in order to run these tests
    return "test_development_idem_{{cookiecutter.service_name}}"


@pytest.fixture(scope="session", name="hub")
def integration_hub(hub):
    hub.pop.sub.add(dyne_name="idem")

    with mock.patch("sys.argv", ["idem", "state"]):
        hub.pop.config.load(["idem", "acct"], "idem", parse_cli=True)

    yield hub


async def _setup_hub(hub):
    # Set up the hub before each function here
    ...


async def _teardown_hub(hub):
    # Clean up the hub after each function here
    ...


@pytest.fixture(scope="function", autouse=True)
async def function_hub_wrapper(hub):
    await _setup_hub(hub)
    yield
    await _teardown_hub(hub)


@pytest.mark.asyncio
@pytest.fixture(scope="module")
async def ctx(hub, acct_subs: List[str], acct_profile: str) -> Dict[str, Any]:
    ctx = NamespaceDict(run_name="test", test=False)

    if not hub.OPT.acct.acct_file:
        pytest.skip(
            "Missing acct_file.  "
            "Add ACCT_FILE to your environment with the path to your encrypted credentials file"
        )
    if not hub.OPT.acct.acct_file:
        pytest.skip(
            "Missing acct_key.  "
            "Add ACCT_KEY to your environment with the fernet key to decrypt your credentials file"
        )

    # Add the profile to the account
    await hub.acct.init.unlock(hub.OPT.acct.acct_file, hub.OPT.acct.acct_key)
    if not hub.acct.UNLOCKED:
        pytest.skip(f"acct could not unlock {hub.OPT.acct.acct_file}")
    ctx.acct = await hub.acct.init.gather(acct_subs, acct_profile)

    # Test if the created ctx is functional; if not then skip all the integration tests that use it
    if not ctx.acct.get("connection"):
        pytest.skip("ctx is not configured correctly")

    yield ctx

    # Close all connections when the tests are complete
    await hub.acct.init.close()


# --------------------------------------------------------------------------------



import pathlib

import pop.hub
from dict_tools.data import NamespaceDict

if __name__ == "__main__":
    root_directory = pathlib.Path.cwd()

    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="pop_create")
    hub.pop.sub.add(dyne_name="cloudspec")
    hub.pop.sub.add(dyne_name="config")
    hub.config.integrate.load(
        ["cloudspec", "pop_create"], "cloudspec", parse_cli=False, logs=False
    )
    ctx = NamespaceDict({{cookiecutter}})

    if ctx.create_plugin == "auto_states":
        hub.cloudspec.init.run(
            ctx,
            root_directory,
            create_plugins=["auto_state", "tool", "tests", "docs"],
        )
    elif ctx.create_plugin == "exec_modules":
        hub.cloudspec.init.run(
            ctx,
            root_directory,
            create_plugins=["exec_modules", "tool", "tests", "docs"],
        )
    elif ctx.create_plugin == "state_modules":
        # For now restrict only CRUDs for exec modules, auto_state already does that
        hub.cloudspec.init.run(
            ctx,
            root_directory,
            create_plugins=["auto_state", "tool", "state_modules", "tests", "docs"],
        )
    elif ctx.create_plugin == "templates":
        hub.cloudspec.init.run(
            ctx,
            root_directory,
            create_plugins=["templates"],
        )
    elif ctx.create_plugin == "test_modules":
        hub.cloudspec.init.run(
            ctx,
            root_directory,
            create_plugins=["templates", "tests"],
        )
    else:
        raise ValueError(f"Invalid input '{ctx.create_plugin}' for --create-plugin.")

    # Sanitize based on other arguments
    if ctx.has_acct_plugin:
        hub.tool.path.rmtree(root_directory / ctx.clean_name / "acct")

    # End with the cicd template
    hub.pop_create.init.run(directory=root_directory, subparsers=["cicd"], **ctx)

    # TODO Run sphinx on the docstrings to make sure it all works



CREATE = """
async def {{function.name}}(hub, ctx{{function.header_params}}, name: str = None) -> Dict[str, Any]:
    r'''
    **Autogenerated function**
    {{function.doc|replace("'" * 3, '"' * 3)}}
    Examples:

        Resource State:

        .. code-block:: sls

            resource_is_present:
              {{ function.ref }}.present:
                - {{ function.required_call_params.split(", ")|join("\n                - ")|replace("=", ": ") }}

        Exec call from the CLI:

        .. code-block:: bash

            idem exec {{ function.ref }}.{{ function.name }} {{ function.required_call_params }}
    '''
"""

UPDATE = """
async def {{function.name}}(hub, ctx{{function.header_params}}, name: str = None, resource_id: str = None, **kwargs) -> Dict[str, Any]:
    r'''
    **Autogenerated function**
    {{function.doc|replace("'" * 3, '"' * 3)}}
    Examples:

        Resource State:

        .. code-block:: sls

            resource_is_present:
              {{ function.ref }}.present:
                - {{ function.required_call_params.split(", ")|join("\n                - ")|replace("=", ": ") }}

        Exec call from the CLI:

        .. code-block:: bash

            idem exec {{ function.ref }}.{{ function.name }} {{ function.required_call_params }}
    '''
"""


DELETE = """
async def {{function.name}}(hub, ctx{{function.header_params}}, name: str = None, resource_id: str = None) -> Dict[str, Any]:
    r'''
    **Autogenerated function**
    {{function.doc|replace("'" * 3, '"' * 3)}}
    Examples:

        Resource State:

        .. code-block:: sls

            resource_is_absent:
              {{ function.ref }}.absent:
                - {{ function.required_call_params.split(", ")|join("\n                - ")|replace("=", ": ") }}

        Exec call from the CLI:

        .. code-block:: bash

            idem exec {{ function.ref }}.{{ function.name }} {{ function.required_call_params }}
    '''
"""

GET = """
async def {{function.name}}(hub, ctx{{function.header_params}}, name: str = None, resource_id: str = None) -> Dict[str, Any]:
    r'''
    **Autogenerated function**
    {{function.doc|replace("'" * 3, '"' * 3)}}

    Examples:

        Unmanaged Resource State:

        .. code-block:: sls

            unmanaged_resource:
              exec.run:
                - path: {{ function.ref }}.{{ function.name }}
                - kwargs:
                  {{ function.required_call_params.split(", ")|join("\n" + " " * 18)|replace("=", ": ") }}

        Exec call from the CLI:

        .. code-block:: bash

            idem exec {{ function.ref }}.{{ function.name }} {{ function.required_call_params }}
    '''
"""

LIST = """
async def list_(hub, ctx{{function.header_params}}) -> Dict[str, Any]:
    r'''
    **Autogenerated function**
    {{function.doc|replace("'" * 3, '"' * 3)}}
    Examples:

        Unmanaged Resource State:

        .. code-block:: sls

            unmanaged_resources:
              exec.run:
                - path: {{ function.ref }}.{{ function.name }}
                - kwargs:
                  {{ function.required_call_params.split(", ")|join("\n" + " " * 18)|replace("=", ": ") }}

        Exec call from the CLI:

        .. code-block:: bash

            idem exec {{ function.ref }}.{{ function.name }} {{ function.required_call_params }}

        Describe call from the CLI:

        .. code-block:: bash

            $ idem describe {{ function.ref }}

    '''
"""



import pathlib
import shutil

import pop.hub
from dict_tools.data import NamespaceDict

if __name__ == "__main__":
    root_directory = pathlib.Path.cwd()

    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="pop_create")
    hub.pop.sub.add(dyne_name="cloudspec")
    hub.pop.sub.add(dyne_name="config")
    hub.config.integrate.load(
        ["cloudspec", "pop_create"], "cloudspec", parse_cli=False, logs=False
    )

    ctx = NamespaceDict({{cookiecutter}})

    # sample files don't matter when creating from a specification
    hub.tool.path.delete(
        root_directory / ctx.clean_name / "exec" / ctx.service_name / "sample.py"
    )
    hub.tool.path.delete(
        root_directory / ctx.clean_name / "states" / ctx.service_name / "sample.py"
    )
    hub.tool.path.delete(
        root_directory
        / ctx.clean_name
        / "autogen"
        / ctx.service_name
        / "templates"
        / "sample.jinja2"
    )

    if ctx.get("templates_dir", None):
        # overwrite all files in target template with source template directory
        target = (
            root_directory / ctx.clean_name / "autogen" / ctx.service_name / "templates"
        )
        # Any existing file will be overwritten
        shutil.copytree(ctx.get("templates_dir"), target, dirs_exist_ok=True)



import base64

from typing import Any
from typing import Dict

from dict_tools.data import NamespaceDict

DEFAULT_ENDPOINT_URL = "{{cookiecutter.servers|first}}"


async def gather(hub, profiles) -> Dict[str, Any]:
    """
    Generate token with basic auth

    Example:
    .. code-block:: yaml

        {{cookiecutter.service_name}}:
          profile_name:
            username: my_user
            password: my_token
            endpoint_url: https://{{cookiecutter.service_name}}.com
    """
    sub_profiles = {}
    for (
            profile,
            ctx,
    ) in profiles.get("{{cookiecutter.service_name}}", {}).items():
        endpoint_url = ctx.get("endpoint_url")

        creds = f"{ctx.get('username')}:{ctx.get('password')}"
        temp_ctx = NamespaceDict(acct={
            "endpoint_url": endpoint_url,
            "headers": {
                "Authorization": f"Basic {base64.b64encode(creds.encode('utf-8')).decode('ascii')}"
            }
        })

        ret = await hub.tool.{{cookiecutter.service_name}}.session.request(
            temp_ctx,
            method="post",
            # TODO: Change Login path
            path="TODO".format(**{}),
            data={},
        )

        if not ret["result"]:
            error = f"Unable to authenticate: {ret.get('comment', '')}"
            hub.log.error(error)
            raise ConnectionError(error)

        # TODO: Find the token value in response
        access_token = ret["ret"][
            "token"
        ]
        sub_profiles[profile] = dict(
            endpoint_url=endpoint_url,
            # TODO: Replace the header that should be passed to future requests
            headers={"Authorization": f"Bearer {access_token}"},
        )
    return sub_profiles



try:
    import inflect

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def __init__(hub):
    hub.tool.format.inflect.ENGINE = inflect.engine()


def singular(hub, string: str) -> str:
    """
    if the string is a plural noun, make it singular.
    Return an empty string if it is already singular.
    """
    if (
        any(string.endswith(s) for s in ("ss", "is", "us", "as", "a"))
        or len(string) < 5
    ):
        return ""
    return hub.tool.format.inflect.ENGINE.singular_noun(string)



import pprint

"""
A sample state using the sample idem_cloud modules for an idea of what a state module should look like.
By convention, all states should have a "present" and an "absent" function, nothing else.
"""


async def present(hub, ctx, name: str, **kwargs):
    """
    Ensure that a resource exists and is in the correct state.

    Example:

    .. code-block:: sls

        my_state_name:
          {{cookiecutter.service_name}}.sample.present:
            name: item_name
            acct_profile: default
            my_kwarg: my_value
    """
    try:
        if name in await hub.exec.{{cookiecutter.service_name}}.sample.list(ctx)["ret"]:
            comment = f"{name} is already exists"
            my_instance = before = await hub.exec.{{cookiecutter.service_name}}.sample.get(ctx, name)["ret"]
        else:
            before = {}
            if ctx.test:
                # If `idem state` was run with the `--test` flag, then don't actually make any changes
                comment = (
                    f"{name} would be created with kwargs: {pprint.pformat(kwargs)}"
                )
            else:
                hub.exec.{{cookiecutter.service_name}}.create(ctx, name, **kwargs)
                comment = f"Created {name} with kwargs: {pprint.pformat(kwargs)}"
            # Now that it exists, retrieve it
            my_instance = await hub.exec.{{cookiecutter.service_name}}.sample.get(ctx, name)["ret"]

        # TODO verify that my_instance is in the correct state and has all the correct parameters
        # TODO This is where the heavy lifting takes place
        # TODO call idem_cloud modules to update the instance as needed.
        # TODO If ctx.test is True, then only comment on what changes would be made

        after = await hub.exec.{{cookiecutter.service_name}}.sample.get(ctx, name)["ret"]
        changes = {"old": before, "new": after}
        result = True
    except Exception as e:
        comment = str(e)
        changes = {}
        result = False

    # This is what is expected of every state return
    return {
        "name": name,
        "changes": changes,
        "result": result,
        "comment": comment,
    }


async def absent(hub, ctx, name: str, **kwargs):

    """
    A sample state using the sample idem_cloud modules for an idea of what an absent function should look like.

    Example:

    .. code-block:: sls

        my_state_name:
          {{cookiecutter.service_name}}.sample.absent:
            name: item_name
            acct_profile: default
            my_kwarg: my_value
    """
    try:
        if name not in await hub.exec.{{cookiecutter.service_name}}.sample.list(ctx)["ret"]:
            comment = f"{name} is already absent"
            before = {}
        else:
            before = await hub.exec.{{cookiecutter.service_name}}.sample.get(ctx, name)["ret"]
            if ctx.test:
                # If `idem state` was run with the `--test` flag, then don't actually make any changes
                comment = (
                    f"{name} would be deleted with kwargs: {pprint.pformat(kwargs)}"
                )
            else:
                hub.exec.{{cookiecutter.service_name}}.delete(ctx, name, **kwargs)["ret"]
                comment = f"Deleted {name} with kwargs: {pprint.pformat(kwargs)}"

        # Some clouds will return an empty dict, some will return an object that is "pending deletion"
        after = await hub.exec.{{cookiecutter.service_name}}.sample.get(ctx, name)["ret"]
        changes = {"old": before, "new": after}
        result = True
    except Exception as e:
        result = False
        changes = {}
        comment = str(e)

    # This is what is expected of every state return
    return {
        "name": name,
        "changes": changes,
        "result": result,
        "comment": comment,
    }



import json
from typing import Any
from typing import Dict

import aiohttp


async def request(
        hub,
        ctx,
        method: str,
        path: str,
        query_params: Dict[str, str] = {},
        data: Dict[str, Any] = {},
        headers: Dict[str, Any] = {},
):
    # Enable trace logging listener for the http client
    trace_config = aiohttp.TraceConfig()
    trace_config.on_request_start.append(hub.tool.{{cookiecutter.service_name}}.session.on_request_start)
    trace_config.on_request_end.append(hub.tool.{{cookiecutter.service_name}}.session.on_request_end)

    # path usually starts with "/" in openapi spec
    url = "".join((ctx.acct.endpoint_url.rstrip("/"), path))

    async with aiohttp.ClientSession(
        loop=hub.pop.Loop, trace_configs=[trace_config],
    ) as session:
        result = dict(ret=None, result=True, status=200, comment=[], headers={})

        if not headers.get("content-type"):
            headers["content-type"] = "application/json"
            headers["accept"] = "application/json"

        if "headers" in ctx.acct:
            # The acct login could set authorization and other headers
            headers.update(ctx.acct.headers)

        query_params_sanitized = {k: v for k, v in query_params.items() if v is not None}
        async with session.request(
            url=url,
            method=method.lower(),
            ssl=False,
            allow_redirects=True,
            params=query_params_sanitized,
            data=json.dumps(data),
            headers=headers,
        ) as response:
            result["status"] = response.status
            result["result"] = 200 <= response.status <= 204
            result["comment"].append(response.reason)
            result["headers"].update(response.headers)
            try:
                result["ret"] = hub.tool.type.dict.namespaced(await response.json())
                response.raise_for_status()
            except Exception as err:
                result["comment"].append(result["ret"])
                result["comment"].append(f"{err.__class__.__name__}: {err}")
                result["result"] = False
                if response.status != 404:
                    try:
                        ret = await response.read()
                        result["ret"] = ret.decode() if hasattr(ret, "decode") else ret
                    except Exception as ex_read_err:
                        result["comment"].append(
                            f"Failed to read response: {ex_read_err.__class__.__name__}: {ex_read_err}")

            return result


async def on_request_start(hub, session, trace_config_ctx, params):
    hub.log.debug("Starting %s" % params)


async def on_request_end(hub, session, trace_config_ctx, params):
    hub.log.debug("Ending %s" % params)



from unittest import mock

import pytest


@pytest.fixture(scope="session", name="hub")
def unit_hub(hub):
    hub.pop.sub.add(dyne_name="idem")
    yield hub


async def _setup_hub(hub):
    # Set up the hub before each function here
    ...


async def _teardown_hub(hub):
    # Clean up the hub after each function here
    ...


@pytest.fixture(scope="function", autouse=True)
async def function_hub_wrapper(hub):
    await _setup_hub(hub)
    yield
    await _teardown_hub(hub)


@pytest.fixture(scope="function")
def mock_hub(hub):
    m_hub = hub.pop.testing.mock_hub()
    m_hub.OPT = mock.MagicMock()
    yield m_hub



import builtins
import keyword


def unclash(hub, string: str) -> str:
    """
    If the string name clashes with a builtin, then append an underscore
    """
    if keyword.iskeyword(string) or string in dir(builtins):
        return f"{string}_"
    return string



async def gather(hub):
    """
    Authenticate to the {{cookiecutter.service_name}} with it's traditional cli tool's config/environment
    """
    sub_profiles = {}

    # TODO add profiles based on traditional {{cookiecutter.service_name}} cli tool auth

    return sub_profiles



import os
import pathlib

import tqdm

from cloudspec import CloudSpec


def run(hub, ctx, root_directory: pathlib.Path):
    if isinstance(root_directory, str):
        root_directory = pathlib.Path(root_directory)
    cloud_spec = CloudSpec(**ctx.cloud_spec)
    tool_dir = root_directory / ctx.clean_name / "tool" / ctx.service_name
    template_dir = (
        root_directory
        / ctx.clean_name
        / "autogen"
        / ctx.service_name
        / "templates"
        / "tool"
    )

    for ref, plugin in cloud_spec.plugins.items():
        mod_file = hub.cloudspec.parse.plugin.touch(tool_dir, ref)
        ref = hub.cloudspec.parse.plugin.ref(ctx, ref)
        module_ref = hub.cloudspec.parse.plugin.mod_ref(ctx, ref, plugin)

        # Set up the base template
        if not plugin.functions:
            to_write = hub.cloudspec.parse.plugin.header(plugin)
        else:
            to_write = hub.cloudspec.parse.plugin.header(plugin)
            for func_name, func_data in tqdm.tqdm(
                plugin.functions.items(), desc=f"{ref} tool functions"
            ):
                if func_name in [
                    "get",
                    "list",
                    "create",
                    "update",
                    "delete",
                    "present",
                    "absent",
                    "describe",
                ] or func_name.startswith("test_"):
                    # These functions are not for tool module
                    continue

                function_request_format_template_file = (
                    f"{template_dir}/{func_name}.jinja2"
                )
                default_request_format_template_file = f"{template_dir}/default.jinja2"
                if os.path.isfile(function_request_format_template_file):
                    with open(function_request_format_template_file, "rb+") as f:
                        request_format = f.read().decode()
                elif os.path.isfile(default_request_format_template_file):
                    with open(default_request_format_template_file, "rb+") as f:
                        request_format = f.read().decode()
                else:
                    # cloudspec was provided with hardcoded request format
                    request_format = cloud_spec.request_format.get(func_name)

                template = hub.tool.jinja.template(
                    f"{hub.cloudspec.template.tool.FUNCTION}\n    {request_format}\n\n"
                )

                try:
                    to_write += template.render(
                        function=dict(
                            name=func_name,
                            ref=ref,
                            module_ref=f"tool.{module_ref}",
                            **func_data,
                            header_params=hub.cloudspec.parse.param.headers(
                                func_data.params
                            ),
                        ),
                        parameter=dict(
                            mapping=hub.cloudspec.parse.param.mappings(func_data.params)
                        ),
                    )
                except Exception as err:
                    hub.log.warning(
                        f"Warning when generating function's action definitions for {func_name}: {err.__class__.__name__}: {err}"
                    )

        mod_file.write_text(to_write)



import dataclasses
from typing import Any
from typing import Dict
from typing import List

import pop.hub
from dict_tools.data import NamespaceDict

HTTP_REQUEST_FORMAT = """
return await hub.tool.{{cookiecutter.service_name}}.session.request(
    method="{{func.hardcoded.method}}",
    path=f"{{param.value.path}}",
    query_params={{param.mapping.query}},
    data={{param.mapping.data}},
)
"""


def create_hub() -> pop.hub.Hub:
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="tool")
    hub.pop.sub.load_subdirs(hub.tool, recurse=True)
    return hub


class CloudSpec(NamespaceDict):
    """
    Validate a cloud spec

    .. code-block:: json

        {
            "project_name": "",
            "service_name": "",
            "api_version": "latest",
            "request_format": "",
            "plugins": {}
        }
    """

    def __init__(
        self,
        project_name: str,
        service_name: str,
        api_version: str = "latest",
        request_format: Dict[str, str] = None,
        plugins: Dict[str, Dict[str, Any]] = None,
    ):
        """
        :param project_name: The name of the project
        :param service_name: The simple clean_name of the cloud
        :param request_format: Map of function call types to the underlying request methods
        :param plugins: A mapping of plugin references on the hub to CloudSpecPlugin jsons
        """
        if request_format is None:
            request_format = {"http": HTTP_REQUEST_FORMAT}
        if plugins:
            validate_plugins = {
                name: CloudSpecPlugin(**plugin_spec)
                for name, plugin_spec in plugins.items()
            }
        else:
            validate_plugins = {}

        super().__init__(
            api_version=api_version,
            project_name=project_name,
            service_name=service_name,
            request_format=request_format,
            plugins=validate_plugins,
        )


class CloudSpecPlugin(NamespaceDict):
    """

    .. code-block:: json

        {
            "{{cookiecutter.service_name}}.sub.sub.plugin_name": {
                "imports": ["import pathlib", "from sys import executable"],
                "virtual_imports": [],
                "func_alias": {"list_":  "list"},
                "virtualname": "",
                "doc": "a module level docstring",
                "functions": {}
            }
        }
    """

    def __init__(
        self,
        doc: str,
        imports: List[str] = None,
        virtual_imports: List[str] = None,
        contracts: List[str] = None,
        func_alias: Dict[str, str] = None,
        virtualname: str = "",
        sub_virtual: bool = None,
        sub_alias: List[str] = None,
        functions: Dict[str, Dict[str, Dict[str, str]]] = None,
        file_vars: Dict[str, Any] = None,
    ):
        """
        Args:
            doc: A docstring for this function
            imports: python imports that will for sure be available, program crashes if not available
            virtual_imports: python imports that prevent the module from loading, plugin does not load if not available
            contracts: A list of contracts to implement for this plugin
            func_alias: A mapping of functions that mirror builtin names, to the names they should have
            virtualname: The name this plugin should be called by on the hub
            sub_virtual: Used to prevent a sub and all it's children from loading
            sub_alias: Only used in init plugins when the parent sub has a name clash
            functions: A mapping of function names to CloudSpecFunction jsons
        """

        if functions:
            validated_functions = {
                name: CloudSpecFunction(**function_spec)
                for name, function_spec in functions.items()
            }
        else:
            validated_functions = {}
        super().__init__(
            doc=doc,
            imports=imports or [],
            virtual_imports=virtual_imports or [],
            contracts=contracts or [],
            func_alias=func_alias or {},
            virtualname=virtualname,
            sub_virtual=sub_virtual,
            sub_alias=sub_alias,
            functions=validated_functions,
            file_vars=file_vars,
        )


class CloudSpecFunction(NamespaceDict):
    """
    .. code-block:: json

        {
            "function_name": {
                "doc": "",
                "return_type": "",
                "hardcoded": {},
                "params": {}
            }
        }
    """

    def __init__(
        self,
        doc: str,
        return_type: str = "None",
        hardcoded: Dict[str, str] = None,
        params: Dict[str, Dict[str, Dict[str, str]]] = None,
    ):
        """
        Args:
            doc: A function docstring
            return_type: return value Type as string
            hardcoded: A mapping of string parameters to string values
            params: A mapping of parameter names to CloudSpecParam jsons
        """
        if params:
            validated_params = {
                name: CloudSpecParam(name=name, **param_spec)
                for name, param_spec in params.items()
            }
        else:
            validated_params = {}
        super().__init__(
            doc=doc,
            return_type=return_type,
            hardcoded=hardcoded or {},
            params=validated_params,
        )


class CloudSpecParam(NamespaceDict):
    """
    .. code-block json::

        {
            "param_name": {
                "doc": "Docstring for this parameter",
                "param_type": "Type",
                "required": True|False,
                "default": "",
                "target": "",
                "target_type": "mapping|value|arg|kwargs"
                "member": {
                    "name": "param_type_name",
                    "params": {
                        "nested_param1_name": CloudSpecParam,
                        "nested_param2_name": CloudSpecParam,
                        ...
                    }
                },
            }
        }
    """

    def __init__(
        self,
        name: str,
        required: bool,
        target: str,
        target_type: str,
        member: dataclasses.make_dataclass(
            cls_name="CloudSpecMember",
            fields=[("name", str), ("params", Dict[str, "CloudSpecParam"])],
        ) = None,
        param_type: str = None,
        doc: str = "",
        default: Any = None,
        snaked: str = None,
    ):
        """
        Args:
            name: Name of the parameter
            doc: Docstring for the parameter
            required: "True|False"
            target: "query|data|path|sdk-class|sdk-subclass|function-call"
            target_type: "mapping|format_str|arg|kwargs"
            param_type: A type hint for this parameter
            member: Nested member params
            default: The default value for this parameter if it is not required
            snaked: Snake case representation of the parameter name to be used, would override generated one
        """
        # If the target_type is "mapping" the target is the name of a dictionary variable
        # If the target_type is "arg" or "kwargs" the target is the name of callable this param should forwarded to
        # If the target_type is "value" then it is directly dropped in and it is included in the function parameters
        if target_type not in ("mapping", "value", "arg", "kwargs"):
            raise TypeError(f"Invalid target type: {target_type}")

        hub = create_hub()

        if member and member.get("params", None):
            validated_member_params = {
                name: CloudSpecParam(name=name, **param_spec)
                for name, param_spec in member.get("params").items()
            }
            member = {"name": member.get("name"), "params": validated_member_params}
        else:
            member = None

        if not snaked:
            # Snake case the parameter every time
            snaked = hub.tool.format.case.snake(name)
            # Add an underscore to the parameter name if it shadows a builtin
            unclashed_name = hub.tool.format.keyword.unclash(snaked) or snaked
            snaked = unclashed_name.replace(" ", "_").replace("-", "_").strip(". ")

        super().__init__(
            doc=doc,
            snaked=snaked,
            param_type=param_type,
            required=required,
            default=default,
            target_type=target_type,
            member=member,
            target=target,
        )



"""Parse parameters related metadata"""
from typing import Dict

import openapi3

__func_alias__ = {"type_": "type"}


def parse_schema_members(hub, params: dict, schema_ref: str, all_schemas: dict):
    """
    Resolve "#/components/schema/Pet type of schema into parameters
    """
    if not schema_ref:
        return

    parameters: list = all_schemas.get(schema_ref)
    if parameters:
        for p in parameters:
            doc = p.get("description")
            nested_member = {}
            if p.get("nested_member", None):
                # parse nested members
                nested_params = {}
                hub.pop_create.openapi3.params.parse_schema_members(
                    nested_params, p.get("nested_member"), all_schemas
                )

                if nested_params.get("primitive", None):
                    param_type = nested_params.get("primitive", {}).get("param_type")
                    doc = nested_params.get("primitive", {}).get("doc")
                else:
                    param_type = hub.pop_create.openapi3.params.type(p.get("type"))
                    nested_member = {
                        "name": p.get("idem_name"),
                        "params": nested_params,
                    }
            else:
                param_type = hub.pop_create.openapi3.params.type(p.get("type"))

            params[p.get("idem_name")] = {
                "required": p.get("required", False),
                "target_type": "mapping",
                "target": "kwargs",
                "param_type": param_type,
                "doc": doc,
            }

            if nested_member:
                params[p.get("idem_name")]["member"] = nested_member


def parse_non_schema_members(hub, parameter: openapi3.paths.Parameter):
    """
    Resolve query, path, header or cookie type of inputs into parameters
    """
    if parameter.in_ == "query":
        target_type = "mapping"
    elif parameter.in_ == "path":
        target_type = "mapping"
    elif parameter.in_ == "header":
        target_type = "mapping"
    elif parameter.in_ == "cookie":
        target_type = "mapping"
    else:
        raise ValueError(f"Unknown parameter type: {parameter.in_}")

    return {
        "required": parameter.required,
        "target_type": target_type,
        "target": parameter.in_,
        "param_type": hub.pop_create.openapi3.params.type(
            parameter.schema.type
            if isinstance(parameter.schema, openapi3.schemas.Schema)
            else None
        ),
        "doc": parameter.description,
    }


def parse_resource_to_request_input_mappings(hub, schema_ref: str, all_schemas: dict):
    resource_to_request_input_mapping = {}
    if not schema_ref:
        return "{}"

    parameters: list = all_schemas.get(schema_ref)
    if parameters:
        for p in parameters:
            if "idem_name" in p and p.get("idem_name"):
                resource_to_request_input_mapping[p.get("idem_name")] = p.get(
                    "actual_name", ""
                )

    # This is used when mapping idem formatted resource attributes to raw request inputs
    # resource_to_request_input_mapping = {
    #     "photo_urls": "photoUrls", -> photoUrls is the attribute on server side but idem name is snaked as photo_urls
    #     "tags": tags,
    # }
    return resource_to_request_input_mapping


def parse_response_mappings(hub, schema_ref: str, all_schemas: dict) -> Dict[str, str]:
    """
    Resolve idem_resource to request payload
    """
    response_mapping = {}
    if not schema_ref:
        return response_mapping

    parameters: list = all_schemas.get(schema_ref)
    if parameters:
        # e.g. a get pet request would get the following response mapping
        # {
        #     "id": "id",
        #     "category": "category",
        #     "name": "name",
        #     "photo_urls": "photoUrls",
        #     "tags": "tags",
        #     "status": "status",
        # }
        for p in parameters:
            if "idem_name" in p and p.get("idem_name"):
                response_mapping[p.get("idem_name")] = p.get("actual_name", "")

    return response_mapping


def type_(hub, param_type: str) -> str:
    if "integer" == param_type:
        return "int"
    elif "boolean" == param_type:
        return "bool"
    elif "number" == param_type:
        return "float"
    elif "string" == param_type:
        return "str"
    elif "array" == param_type:
        return "list"
    elif "object" == param_type:
        return "dict"
    elif "nested" == param_type:
        # This translates into makedataclass
        return "{}"
    elif "nested_array" == param_type:
        # This translates into List[makedataclass]
        return "List[{}]"
    else:
        return ""



import pathlib


def run(hub, ctx, root_directory: pathlib.Path):
    """
    This is purposefully left as no-op. When used, no code will be rendered but default templates will be provided.
    This is only applicable for "swagger" or "openapi3" specifications.
    """



import json
import pathlib
import timeit
from typing import List

from cloudspec import CloudSpec


def __init__(hub):
    hub.pop.sub.add(dyne_name="tool")
    hub.pop.sub.load_subdirs(hub.tool, recurse=True)
    hub.pop.sub.load_subdirs(hub.cloudspec, recurse=True)


def cli(hub):
    """
    Validate json from the cli
    """
    hub.pop.sub.add(dyne_name="output")
    hub.pop.config.load(["cloudspec", "rend", "pop_create"], cli="cloudspec")

    with open(hub.OPT.cloudspec.input_file, "w+") as fh:
        data = json.load(fh)

    if hub.SUBPARSER == "validate":
        validated_spec = CloudSpec(**data)
        hub.output[hub.OPT.rend.output].display(dict(validated_spec))
    elif hub.SUBPARSER == "create":
        root_directory = pathlib.Path(hub.OPT.pop_create.directory)
        ctx = hub.pop_create.idem_cloud.init.context(
            hub.pop_create.init.context(), root_directory
        )
        ctx.cloud_spec = data
        hub.cloudspec.init.run(ctx, root_directory, hub.OPT.cloudspec.create_plugins)
    else:
        hub.log.error(f"Unknown subparser: {hub.SUBPARSER}")


def run(
    hub,
    ctx,
    root_directory: pathlib.Path,
    create_plugins: List[str],
):
    start_time = timeit.default_timer()
    for create_plugin in create_plugins:
        try:
            hub.log.info(f"Running create plugin: {create_plugin}")
            hub.cloudspec.create[create_plugin].run(ctx, root_directory)
        except Exception as e:
            hub.log.error(f"Failed to run create plugin: {create_plugin}")
            hub.log.error(e)
            raise
    print(f"Total Duration: {timeit.default_timer() - start_time} seconds")



import openapi3


def parse(hub, api: openapi3.OpenAPI):
    schemas = {}

    # Sometimes a schema could just point to different schema object
    # Let's hold them into a separate mapping for adding them into flat schema structure
    transitive_schema_mappings = {}
    # Get a flat structure for schemas
    for name, schema in api.components.raw_element.items():
        if name == "schemas":
            for object_name, param in schema.items():
                parameters = []
                if param.get("$ref", None):
                    # We may need to sanitize it later
                    # "InventoryOrder": {
                    #     "$ref": "#/components/schemas/StoreOrder"
                    # },
                    transitive_schema_mappings[
                        f"#/components/schemas/{object_name}"
                    ] = param.get("$ref")
                elif param.get("type", "") == "object":
                    # "StoreOrder": {
                    #     "type": "object",
                    #     "properties": {
                    #         "name": {
                    #             "type": "string",
                    #             "description": ""
                    #         },
                    #     }
                    for prop_name, prop_value in param.get("properties", {}).items():
                        # Check for nested members first
                        # "pet_order": {
                        #    "$ref": "#/components/schemas/StoreOrder"
                        #  },
                        nested_member_ref = prop_value.get("$ref")
                        nested_member_ref_type = "nested" if nested_member_ref else None
                        if "array" == prop_value.get("type"):
                            # "tags": {
                            #     "type": "array",
                            #     "items": {
                            #         "$ref": "#/definitions/Tag"
                            #     }
                            # },
                            # "photoUrls": {
                            #     "type": "array",
                            #     "items": {
                            #         "type": "string",
                            #     }
                            # },
                            nested_member_ref = prop_value.get("items", {}).get("$ref")
                            # Keeping the default array for now so that it is becomes a list in function params
                            nested_member_ref_type = (
                                "nested_array" if nested_member_ref else "array"
                            )

                        parameters.append(
                            dict(
                                idem_name=hub.tool.format.case.snake(prop_name),
                                actual_name=prop_name,
                                required=prop_name in param.get("required", []),
                                type=nested_member_ref_type
                                if nested_member_ref_type
                                else prop_value.get("type"),
                                nested_member=nested_member_ref,
                                description=prop_value.get("description") or prop_name,
                            )
                        )
                else:
                    # We may need to sanitize it later
                    # "InventoryOrderStatus": {
                    #     "type": "string",
                    #     "description": "<>",
                    # },
                    parameters.append(
                        dict(
                            idem_name="primitive",
                            type=param.get("type"),
                            description=param.get("description"),
                        )
                    )

                schemas[f"#/components/schemas/{object_name}"] = parameters

    for key, actual_schema_ref in transitive_schema_mappings.items():
        # Update schemas with actual schema reference
        schemas[key] = schemas[actual_schema_ref]

    return schemas



import textwrap3


def indent(hub, data: str, level: str = 1) -> str:
    return textwrap3.indent(data, "    " * level)


def wrap(hub, data: str, width: int = 80) -> str:
    return textwrap3.wrap(data, width)



import json
import pathlib

from dict_tools.data import NamespaceDict


def __init__(hub):
    hub.pop.sub.add(dyne_name="cloudspec")


def context(hub, ctx, directory: pathlib.Path):
    # all template files are copied without render
    ctx["_copy_without_render"] = ["*.jinja2"]

    # non openapi/swagger based would not have cloud_spec here
    if "cloud_spec" not in ctx or not ctx.cloud_spec:
        # If an acct plugin was passed in then we don't need to create auth plugins
        if ctx.get("simple_service_name"):
            ctx.service_name = ctx.simple_service_name
        elif not ctx.get("service_name"):
            ctx.service_name = (
                ctx.clean_name.replace("idem", "").replace("cloud", "").strip("_")
            )

        ctx.clean_api_version = ctx.get("clean_api_version")
        ctx.cloud_api_version = ctx.get("cloud_api_version")
        ctx.servers = ctx.get("servers", [""])
        if ctx.specification:
            with open(ctx.specification, "w+") as fh:
                data = json.load(fh)

            ctx.cloud_spec = data
        else:
            ctx.cloud_spec = NamespaceDict(
                api_version=ctx.get("cloud_api_version"),
                project_name=ctx.project_name,
                service_name=ctx.service_name,
                request_format=None,
                plugins={},
            )

        ctx.has_acct_plugin = bool(ctx.acct_plugin)
        if not ctx.has_acct_plugin:
            # Create auth plugins
            ctx.acct_plugin = ctx.service_name

    return ctx



FUNCTION = """
async def {{function.name}}(hub, ctx{{function.header_params}})  -> Dict[str, Any]:
    r'''
    **Autogenerated function**
    {{function.doc|replace("'" * 3, '"' * 3)}}
    Examples:

        .. code-block:: bash

            idem exec {{ function.ref }}.{{ function.name }} {{ function.required_call_params }}

        .. code-block:: python

            async def my_func(hub, ctx):
                ret = await hub.exec.{{ function.ref }}.{{ function.name }}(ctx, {{ function.required_call_params }})
                assert ret.result is True, ret.ref + ": " + ret.comment
                return ret.ret
    '''
"""



"""
Functions for processing plugins
"""
import pathlib

from cloudspec import CloudSpecPlugin


def header(hub, plugin: CloudSpecPlugin) -> str:
    """
    Initialize the render of a plugin file and return the template
    """
    # noinspection JinjaAutoinspect
    template = hub.tool.jinja.template(hub.cloudspec.template.plugin.HEADER)

    return template.render(plugin=plugin)


def ref(hub, ctx, ref: str) -> str:
    split = ref.split(".")
    subs = split[:-1]
    mod = split[-1]
    return ".".join([ctx.service_name] + subs + [mod])


def mod_ref(hub, ctx, ref: str, plugin: CloudSpecPlugin) -> str:
    split = ref.split(".")
    subs = split[:-1]
    mod = split[-1]
    return ".".join([ctx.service_name] + subs + [plugin.virtualname or mod])


def touch(hub, root: pathlib.Path, ref: str):
    """
    Create all the files underneath the new sub
    """
    split = ref.split(".")
    subs = split[:-1]
    mod = split[-1]

    for sub in subs:
        root = root / sub
    root = root / f"{mod}.py"
    hub.tool.path.touch(root)
    return root



from typing import Any
from typing import Dict

from dict_tools.data import NamespaceDict

DEFAULT_ENDPOINT_URL = "{{cookiecutter.servers|first}}"

async def gather(hub, profiles) -> Dict[str, Any]:
    """
    Authenticate with api_key

    Example:
    .. code-block:: yaml

        {{cookiecutter.service_name}}:
          profile_name:
            api_version: <version-info>
            api_key: <api-key>
            endpoint_url: {{cookiecutter.servers|join('|')}}
    """
    sub_profiles = {}
    for (
        profile,
        ctx,
    ) in profiles.get("{{cookiecutter.service_name}}", {}).items():
        api_key = ctx.get("api_key")
        endpoint_url = ctx.get("endpoint_url")
        api_version = ctx.get("api_version")

        temp_ctx = NamespaceDict(acct={
            "endpoint_url": endpoint_url,
            "api_version": api_version,
        })

        ret = await hub.tool.{{cookiecutter.service_name}}.session.request(
            temp_ctx,
            method="post",
            # TODO: Change login path
            path="TODO".format(**{}),
            data={
                # TODO: pass api_key to respective
                "X-API-KEY": api_key,
                "refreshToken": api_key
            },
        )

        if not ret["result"]:
            error = f"Unable to authenticate: {ret.get('comment', '')}"
            hub.log.error(error)
            raise ConnectionError(error)

        access_token = ret["ret"][
            # TODO: Replace response key with corresponding API response
            "token"
        ]
        sub_profiles[profile] = dict(
            endpoint_url=endpoint_url,
            headers={"Authorization": f"Bearer {access_token}"},
        )
    return sub_profiles



import pathlib

import requests


def context(hub, ctx, directory: pathlib.Path):
    openapi_specs = []
    for specification in ctx.specification:
        swagger = hub.pop_create.openapi3.init.read(source=specification)

        # Convert swagger to openapi3
        response = requests.post(
            "https://converter.swagger.io/api/convert",
            json=dict(swagger),
            headers={"Content-Type": "application/json"},
        )
        openapi_specs.append(response.json())

    ctx.specification = openapi_specs

    # get ctx from openapi3 with changes that turn swagger into openapi3
    hub.pop_create.init.run(directory=directory, subparsers=["openapi3"], **ctx)

    return ctx



import base64

DEFAULT_ENDPOINT_URL = "{{cookiecutter.servers|first}}"


async def gather(hub):
    """
    Authenticate with a username and password to the given endpoint url.
    Any extra parameters will be saved as part of the profile.

    Example:

    .. code-block:: sls

        {{cookiecutter.service_name}}:
          profile_name:
            username: my_user
            password: my_token
            endpoint_url: https://{{cookiecutter.service_name}}.com
    """
    sub_profiles = {}
    for profile, ctx in hub.acct.PROFILES.get(
        "{{cookiecutter.service_name}}.basic_auth", {}
    ).items():
        creds = f"{ctx.pop('username')}:{ctx.pop('password')}"
        sub_profiles[profile] = dict(
            endpoint_url=ctx.pop("endpoint_url"),
            headers={
                "Authorization": f"Basic {base64.b64encode(creds.encode('utf-8')).decode('ascii')}"
            },
            **ctx
        )

    return sub_profiles



"""
Templates for plugins
"""

HEADER = '''
"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

{{ plugin.doc }}
"""

{% if plugin.imports -%}
{{ plugin.imports|join('\n') }}
{%- endif %}

{% if plugin.virtual_imports -%}
try:
    {{ plugin.virtual_imports|join('\n') }}
    HAS_LIBS = True,
except ImportError as e:
    HAS_LIBS = False, str(e)
{%- endif %}

{%- if plugin.contracts -%}
__contracts__ = {{plugin.contracts}}
{%- endif %}

{%- if plugin.virtual_imports -%}
def __virtual__(hub):
    return HAS_LIBS
{%- endif %}

{%- if plugin.virtualname -%}
__virtualname__ = "{{plugin.virtualname}}"
{%- endif %}

{%- if plugin.sub_alias -%}
__sub_alias__ = {{plugin.sub_alias}}
{%- endif %}

{% if plugin.func_alias -%}
__func_alias__ = {{plugin.func_alias}}
{% endif %}

{% if plugin.file_vars %}
{% for var, val in plugin.file_vars.items() %}
{{ var }} = {{ val }}
{% endfor %}
{% endif %}
'''



import os.path
import pathlib

import tqdm

from cloudspec import CloudSpec


def run(hub, ctx, root_directory: pathlib.Path or str):
    if isinstance(root_directory, str):
        root_directory = pathlib.Path(root_directory)
    cloud_spec = CloudSpec(**ctx.cloud_spec)
    exec_dir = root_directory / ctx.clean_name / "exec" / ctx.service_name
    template_dir = (
        root_directory
        / ctx.clean_name
        / "autogen"
        / ctx.service_name
        / "templates"
        / "exec"
    )

    for ref, plugin in cloud_spec.plugins.items():
        mod_file = hub.cloudspec.parse.plugin.touch(exec_dir, ref)
        ref = hub.cloudspec.parse.plugin.ref(ctx, ref)

        plugin["func_alias"] = {"list_": "list"}
        if not plugin.get("contracts"):
            # plugin["contracts"] = ["auto_state", "soft_fail"]
            # At the moment, we do not fully support auto_state
            plugin["contracts"] = ["soft_fail"]

        exec_ref = hub.cloudspec.parse.plugin.mod_ref(ctx, ref, plugin)

        # Set up the base template
        if not plugin.functions:
            to_write = hub.cloudspec.parse.plugin.header(plugin)

        else:
            to_write = hub.cloudspec.parse.plugin.header(plugin)
            mod_file.write_text(to_write)
            func_data = hub.cloudspec.parse.function.parse(
                plugin.functions, targets=("get", "list", "create", "update", "delete")
            )

            present_parameter = {}
            if plugin.functions.get("create"):
                present_parameter = hub.cloudspec.parse.param.simple_map(
                    plugin.functions["create"].params
                )

            # Make the get, list, create, delete, and update functions; these are required for every auto_state exec module
            for function_name in tqdm.tqdm(
                ["get", "list", "create", "update", "delete"],
                desc=f"{ref} auto_state functions",
            ):
                if func_data.get(function_name):
                    base_template = hub.cloudspec.template.auto_state[
                        function_name.upper()
                    ]

                    request_format_template_file = (
                        f"{template_dir}/{function_name}.jinja2"
                    )
                    if os.path.isfile(request_format_template_file):
                        with open(f"{template_dir}/{function_name}.jinja2", "rb+") as f:
                            request_format = f.read().decode()
                    else:
                        request_format = cloud_spec.request_format.get(function_name)

                    template_str = f"{base_template}\n    {request_format}\n\n\n"
                    template = hub.tool.jinja.template(template_str)

                    to_write += template.render(
                        service_name=cloud_spec.service_name,
                        function=dict(
                            ref=ref,
                            exec_ref=f"exec.{exec_ref}",
                            **func_data[function_name]["function"],
                        ),
                        parameter=func_data[function_name]["parameter"],
                        present_parameter=present_parameter,
                    )
        mod_file.write_text(to_write)



FUNCTION = """
async def {{function.name}}(
    hub,
    ctx{{function.header_params}}
) -> Dict[str, Any]:
    r\"""
    **Autogenerated function**
    {{function.doc|replace("'" * 3, '"' * 3)}}
    \"""
"""



CLI_CONFIG = {
    "create_plugin": {"subcommands": ["_global_"], "dyne": "pop_create"},
    "simple_service_name": {
        "options": ["--cloud", "--cloud-name"],
        "subcommands": ["_global_"],
        "dyne": "pop_create",
    },
    "acct_plugin": {
        "subcommands": ["idem-cloud", "openapi3", "swagger"],
        "dyne": "pop_create",
    },
    "specification": {
        "options": ["--spec", "--file", "--url"],
        "nargs": "+",
        "subcommands": ["openapi3", "swagger", "idem-cloud"],
        "dyne": "pop_create",
    },
    "templates_dir": {
        "options": ["--td"],
        "subcommands": ["_global_"],
        "dyne": "pop_create",
    },
}
CONFIG = {
    "create_plugin": {
        "default": "state_modules",
        "dyne": "pop_create",
        "help": "The `create` plugin to use for resource modules. The other options are 'auto_states', 'exec_modules'.",
    },
    "acct_plugin": {
        "default": None,
        "help": "The acct plugin to use for authentication -- default is to create a new plugin",
        "dyne": "pop_create",
    },
    "simple_service_name": {
        "default": None,
        "help": "Short name of the cloud being bootstrapped",
        "dyne": "pop_create",
    },
    "specification": {
        "default": None,
        "help": "The url or file path to a spec",
        "dyne": "pop_create",
    },
    "templates_dir": {
        "default": None,
        "help": "The full absolute path of directory where jinja templates are saved",
        "dyne": "pop_create",
    },
}
SUBCOMMANDS = {
    # https://openapi.tools/#converters
    "idem-cloud": {"help": "Boostrap an idem cloud project", "dyne": "pop_create"},
    "openapi3": {
        "help": "Create idem_cloud modules based off of an openapi3 spec",
        "dyne": "pop_create",
    },
    "swagger": {
        "help": "Create idem_cloud modules based off of a swagger spec",
        "dyne": "pop_create",
    },
}
DYNE = {
    "pop_create": ["pop_create"],
    "cloudspec": ["cloudspec"],
    "tool": ["tool"],
    "jinja": ["jinja"],
}
